//
//  QUQRCodeScannerViewController.swift
//  qu_qr_code_scanner
//
//  Created by edy on 2021/7/15.
//

import AVFoundation
import UIKit

class QUQRCodeScannerViewController: UIViewController, AVCaptureMetadataOutputObjectsDelegate {
    var resultBlock: ((_ result: String) -> Void)?

    private var scanRectView = UIView() // 特效活动区域
    private var captureSession: AVCaptureSession!
    private var previewLayer: AVCaptureVideoPreviewLayer!

    private var scanImg = UIImageView(image: QUImagesHelper.imageFor(name: "scan"))
    private var choseImgBtn = UIButton()
    private let maxWindowSize = max(UIScreen.main.bounds.size.width, UIScreen.main.bounds.size.height)
    private let minWindowSize = min(UIScreen.main.bounds.size.width, UIScreen.main.bounds.size.height)
    private lazy var scanSize = CGSize(width: minWindowSize * 3 / 4, height: minWindowSize * 3 / 4)

    private var isEnLarge = true

    override func viewWillAppear(_ animated: Bool) {
        super.viewWillAppear(animated)
        startAnimation()

        ifDenied()

        if captureSession?.isRunning == false {
            captureSession.startRunning()
        }
    }

    override func viewWillDisappear(_ animated: Bool) {
        super.viewWillDisappear(animated)

        if captureSession?.isRunning == true {
            captureSession.stopRunning()
        }
    }

    override func viewDidLoad() {
        super.viewDidLoad()

        initView()
        setScanner()
        orientationDidChangeNotification()
    }

    func initView() {
        let backImg = QUImagesHelper.imageFor(name: "back")
        let backBtn = UIBarButtonItem(image: backImg, style: .plain, target: self, action: #selector(backAction))
        backBtn.tintColor = UIColor.white
        navigationItem.setLeftBarButton(backBtn, animated: false)

        let img = QUImagesHelper.imageFor(name: "photo")
        choseImgBtn.setImage(img, for: .normal)
        choseImgBtn.addTarget(self, action: #selector(chooseImg), for: .touchUpInside)

        view.isUserInteractionEnabled = true
        let tapGesture = UITapGestureRecognizer(target: self, action: #selector(tapGesture))
        tapGesture.numberOfTapsRequired = 2
        view.addGestureRecognizer(tapGesture)
    }

    @objc func tapGesture() {
        CATransaction.begin()
        CATransaction.setAnimationDuration(0.5)
        if isEnLarge {
            previewLayer.transform = CATransform3DMakeScale(3, 3, 3)
        } else {
            previewLayer.transform = CATransform3DMakeScale(1, 1, 1)
        }
        CATransaction.commit()
        isEnLarge = !isEnLarge
    }

    @objc
    private func backAction() {
        navigationController?.dismiss(animated: true, completion: nil)
    }

    @objc func chooseImg() {
        let pick = UIImagePickerController()
        pick.delegate = self
        pick.sourceType = .photoLibrary
        pick.modalPresentationStyle = .fullScreen
        present(pick, animated: true, completion: nil)
    }

    func metadataOutput(_ output: AVCaptureMetadataOutput, didOutput metadataObjects: [AVMetadataObject], from connection: AVCaptureConnection) {
        captureSession.stopRunning()

        if let metadataObject = metadataObjects.first {
            guard let readableObject = metadataObject as? AVMetadataMachineReadableCodeObject else { return }
            guard let stringValue = readableObject.stringValue else { return }
            AudioServicesPlaySystemSound(SystemSoundID(kSystemSoundID_Vibrate))
            found(code: stringValue)
        }

        dismiss(animated: true)
    }

    func found(code: String) {
        if let b = resultBlock {
            print("二维码结果：", code)
            b(code)
        }
    }

    func startAnimation() {
        let animation = CABasicAnimation(keyPath: "transform.translation.y")
        animation.fromValue = 0
        animation.toValue = scanSize.height - 30
        animation.duration = 2
        animation.repeatCount = 99999
        scanImg.layer.add(animation, forKey: nil)
    }

    private func orientationDidChangeNotification() {
        let name = UIDevice.orientationDidChangeNotification
        NotificationCenter.default.addObserver(self, selector: #selector(orientationDidChangeHandle), name: name, object: nil)
    }

    @objc
    private func orientationDidChangeHandle() {
        DispatchQueue.main.asyncAfter(deadline: .now() + 0.1) {
            self.updateScannerRect()
        }
    }

    private func setScanner() {
        view.backgroundColor = UIColor.black
        captureSession = AVCaptureSession()

        guard let videoCaptureDevice = AVCaptureDevice.default(for: .video) else { return }
        let videoInput: AVCaptureDeviceInput

        do {
            videoInput = try AVCaptureDeviceInput(device: videoCaptureDevice)
        } catch {
            return
        }

        if captureSession.canAddInput(videoInput) {
            captureSession.addInput(videoInput)
        } else {
            failed()
            return
        }

        let metadataOutput = AVCaptureMetadataOutput()

        if captureSession.canAddOutput(metadataOutput) {
            captureSession.addOutput(metadataOutput)

            metadataOutput.setMetadataObjectsDelegate(self, queue: DispatchQueue.main)
            metadataOutput.metadataObjectTypes = [.qr]
        } else {
            failed()
            return
        }

        previewLayer = AVCaptureVideoPreviewLayer(session: captureSession)
        previewLayer.frame = view.layer.bounds
        previewLayer.videoGravity = .resizeAspectFill
        previewLayer.backgroundColor = UIColor.black.withAlphaComponent(0.75).cgColor
        view.layer.addSublayer(previewLayer)
        view.addSubview(choseImgBtn)
        view.addSubview(scanRectView)
        scanRectView.addSubview(scanImg)
        updateScannerRect()
    }

    func failed() {
        let ac = UIAlertController(title: "不支持扫码", message: "您的设备没有摄像头，请使用摄像头功能完整的设备.", preferredStyle: .alert)
        ac.addAction(UIAlertAction(title: "确定", style: .default))
        present(ac, animated: true)
        captureSession = nil
    }

    private func updateScannerRect() {
        startAnimation()
        let viewSize = view.frame.size
        previewLayer.frame = view.bounds
        choseImgBtn.frame = CGRect(x: viewSize.width - 80, y: viewSize.height - 100, width: 60, height: 60)

        if maxWindowSize < 500 {
            captureSession.sessionPreset = AVCaptureSession.Preset.vga640x480
        } else {
            captureSession.sessionPreset = AVCaptureSession.Preset.high
        }

        scanRectView.frame = CGRect(x: 0, y: 0, width: minWindowSize,
                                    height: minWindowSize)
        scanRectView.center = CGPoint(x: UIScreen.main.bounds.midX,
                                      y: UIScreen.main.bounds.midY)
        scanImg.frame = CGRect(x: 0, y: 0, width: minWindowSize, height: 30)
        scanImg.contentMode = .scaleAspectFit

        let orientation = UIDevice.current.orientation
        if let connection = previewLayer.connection {
            if connection.isVideoOrientationSupported,
               let videoOrientation = AVCaptureVideoOrientation(rawValue: orientation.rawValue)
            {
                connection.videoOrientation = videoOrientation
            }
        }
    }

    func ifDenied() {
        // First we check if the device has a camera (otherwise will crash in Simulator - also, some iPod touch models do not have a camera).
        if UIImagePickerController.isSourceTypeAvailable(.camera) {
            let authStatus = AVCaptureDevice.authorizationStatus(for: AVMediaType.video)
            switch authStatus {
            case .authorized:
                break
            case .denied:
                alertPromptToAllowCameraAccessViaSettings()
            case .notDetermined:
                break
            default:
                break
            }
        } else {
            let alertController = UIAlertController(title: "相机错误", message: "摄像头不存在", preferredStyle: .alert)
            let defaultAction = UIAlertAction(title: "确定", style: .default, handler: { _ in
            })
            alertController.addAction(defaultAction)
            present(alertController, animated: true, completion: nil)
        }
    }

    func alertPromptToAllowCameraAccessViaSettings() {
        let alert = UIAlertController(title: "需要提供摄像头权限扫码", message: "请前往设置打开该权限", preferredStyle: .alert)
        alert.addAction(UIAlertAction(title: "前往设置", style: .cancel) { _ in
            if let appSettingsURL = URL(string: UIApplication.openSettingsURLString) {
                UIApplication.shared.open(appSettingsURL, options: [:], completionHandler: nil)
            }
        })
        present(alert, animated: true, completion: nil)
    }
}

extension QUQRCodeScannerViewController: UIImagePickerControllerDelegate, UINavigationControllerDelegate {
    func imagePickerController(_ picker: UIImagePickerController, didFinishPickingMediaWithInfo info: [UIImagePickerController.InfoKey: Any]) {
        picker.dismiss(animated: true)
        var image: UIImage?

        image = info[.originalImage] as? UIImage
        guard let image = image else {
            print("图片解析出错")
            return
        }
        let ciImage = CIImage(image: image)!

        let context = CIContext(options: nil)
        let detector = CIDetector(ofType: CIDetectorTypeQRCode, context: context, options: [CIDetectorAccuracy: CIDetectorAccuracyHigh])

        let features = detector?.features(in: ciImage)
        for feature in features as! [CIQRCodeFeature] {
            let result = feature.messageString ?? ""
            found(code: result)
            dismiss(animated: true)
        }
    }
}
